/*
 * Game logic and command parsing.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "abalone.h"

#undef ANSI /* Use VT colour sequences for trace readability */
#if defined(ANSI)
# define NORMAL	"[m"
# define RED	"[31;47m"
# define BLUE	"[34;47m"
# define GREY	"[;37m"
# define RGREY	"[37;47m"
# define HYPHEN	"\x12"
#else
# define NORMAL	""
# define RED	""
# define BLUE	""
# define GREY	""
# define RGREY	""
# define HYPHEN	"-"
#endif

#define INVALID	((coord_t)-1)
#define ROW(c)	((c) >= 'a' && (c) <= 'i' ? (c) - 'a' : \
		 (c) >= 'A' && (c) <= 'I' ? (c) - 'A' : INVALID)
#define COL(c)	((c) >= '1' && (c) <= '9' ? (c) - '1' : INVALID)
#define abs(x)	((x) < 0 ? -(x) : (x))
#define sgn(x)	((x) == 0 ? 0 : (x) < 0 ? -1 : 1)

const char *abalone_error = NULL;

/* Conventional starting position */
static abalone_cells conventional = 
{
              { 2,2,2,2,2, 0,0,0,0 },
	     { 2,2,2,2,2,2, 0,0,0 },
	    { 0,0,2,2,2,0,0, 0,0 },
	   { 0,0,0,0,0,0,0,0, 0 },
	 {  0,0,0,0,0,0,0,0,0  },
	{ 0, 0,0,0,0,0,0,0,0 },
       { 0,0, 0,0,1,1,1,0,0 },
      { 0,0,0, 1,1,1,1,1,1 },
     { 0,0,0,0, 1,1,1,1,1 }
};

/* Construct a new board, initialising it to the conventional start positions */
struct abalone *
new_abalone()
{
	struct abalone *a;

	a = (struct abalone *)malloc(sizeof *a);
	memcpy(a->cell, conventional, sizeof a->cell);
	a->lostpieces[1] = 0;
	a->lostpieces[2] = 0;
	a->refcnt = 0;
	a->previous = NULL;
	a->prevplayer = 0;
	memset(a->prevcmd, 0, sizeof a->prevcmd);
	return a;
}

/* Create a move result state */
static struct abalone *
new_abalone_move(a0, player, cmd)
	struct abalone *a0;
	player_t player;
	char *cmd;
{
	struct abalone *a;

	a = (struct abalone *)malloc(sizeof *a);
	memcpy(a->cell, a0->cell, sizeof a->cell);
	a->lostpieces[1] = a0->lostpieces[1];
	a->lostpieces[2] = a0->lostpieces[2];
	a->refcnt = 0;
	a->previous = a0;
	a->prevplayer = player;
	a0->refcnt++;
	strncpy(a->prevcmd, cmd, sizeof a->prevcmd);
	return a;
}

/* Release storage allocated for a board */
void
free_abalone(a)
	struct abalone *a;
{
	if (a->refcnt == 0) {
		if (a->previous != NULL)
			free_abalone(a->previous);
		free(a);
	} else {
		a->refcnt--;
	}
}

/* Return true if the cell coordinate is off the board */
int
isoffboard(r, c)
	coord_t r, c;
{
	return (r <=4 && (c < 0 || c > r + 4)) || 
	       (r > 4 && (c < r - 4 || c > 8)) ||
	       r < 0 || r > 8;
}

/* Return true if the two coordinates are adjacent */
static int
isadjacent(r1, c1, r2, c2)
	coord_t r1,c1,r2,c2;
{
	return	(r1 == r2 && (c1 == c2 - 1 || c1 == c2 + 1)) ||
		(c1 == c2 && (r1 == r2 - 1 || r1 == r2 + 1)) ||
		(r1 == r2 - 1 && c1 == c2 - 1) ||
		(r1 == r2 + 1 && c1 == c2 + 1);
}

/* Print out the current board */
void
abalone_print(a)
	struct abalone *a;
{
	coord_t r,c;

	printf(GREY "         1 2 3 4 5" NORMAL "\n");
	printf(GREY "        / / / / / 6" NORMAL "\n");
	for (r = 0; r < 9; r++) {
	    if (r < 4) printf("%*s", 4-r, "");
	    if (r > 4) printf("%*s", r-4, "");
	    printf(GREY "%c- ", r+'a');
	    for (c = 0; c < 9; c++) {
		if (isoffboard(r, c))
			continue;
		if (c > (r < 4 ? 0 : r - 4))
			printf(GREY HYPHEN);
		switch (a->cell[r][c]) {
		case 1:
			printf(BLUE "1");
			break;
		case 2:
			printf(RED "2");
			break;
		default:
			printf(RGREY ".");
		}
	    }
	    if (r == 3) 
		printf(GREY " /" NORMAL);
	    else if (r < 3)
		printf(GREY " / %c" NORMAL, '7'+r);
	    else
		printf("%s", NORMAL);

	    if (r == 2 && a->lostpieces[1]) {
		printf("  ");
		for (c = 0; c < a->lostpieces[1]; c++)
		    printf(" " BLUE "1" NORMAL);
	    }
	    if (r == 6 && a->lostpieces[2]) {
		printf("      ");
		for (c = 0; c < a->lostpieces[2]; c++)
		    printf(" " RED "2" NORMAL);
	    }
	    printf("\n");
	}
        printf("\n");
}

#define AERROR(m) do { abalone_error=m; return NULL; } while(0)

/* Parse and perform a move, returning NULL on success, or an error message */
struct abalone *
abalone_move(a, player, cmd)
	struct abalone *a;
	player_t player;
	const char *cmd;
{
	coord_t r1,c1, r2,c2, r3,c3;
	int mylen, theirlen;
	coord_t ldr, ldc, mdr, mdc, r, c;
	int i;
	struct abalone *b;
	struct abalone *result;

	if (a->previous == NULL && player != 1) 
		AERROR("Player 1 must move first");

	if (a->previous != NULL && a->prevplayer != OTHER(player))
		AERROR("Players must take turn");

	if (strcmp(cmd, "pass") == 0) {
		result = new_abalone_move(a, player, "pass");
		goto check;
	}

	if (strlen(cmd) == 5 && cmd[2] == ',') {
		/* xx,yy */
		r1 = r2 = ROW(cmd[0]);
		c1 = c2 = COL(cmd[1]);
		r3 = ROW(cmd[3]);
		c3 = COL(cmd[4]);
	} else if (strlen(cmd) == 8 && cmd[2] == '-' && cmd[5] == ',') {
		/* xx-yy,zz */
		r1 = ROW(cmd[0]);
		c1 = COL(cmd[1]);
		r2 = ROW(cmd[3]);
		c2 = COL(cmd[4]);
		r3 = ROW(cmd[6]);
		c3 = COL(cmd[7]);
	} else
		AERROR("malformed command");

	if (r1 == INVALID || r2 == INVALID || r3 == INVALID ||
	    c1 == INVALID || c2 == INVALID || c3 == INVALID)
		AERROR("malformed position");

	/* Normalize r3c3 to be adjacent to r1c1 if it's not aldready */
	if (!isadjacent(r1,c1,r3,c3)) {
		if (!isadjacent(r2,c2,r3,c3))
			AERROR("invalid destination");
		r3 += r1 - r2;
		c3 += c1 - c2;
	}

	/* Ensure r1c1-r2c2 is a straight line */
	if (r1 == r2) {
		mylen = abs(c1 - c2) + 1;
	} else if (c1 == c2) {
		mylen = abs(r1 - r2) + 1;
	} else if (abs(r1 - r2) == abs(c1 - c2)) {
		mylen = abs(r1 - r2) + 1;
	} else
		AERROR("invalid line");

	if (mylen == 0)
		AERROR("empty line");
	if (mylen > 3)
		AERROR("trying to move too many pieces");

	/* Compute the movement vector */
	mdr = sgn(r3 - r1); mdc = sgn(c3 - c1);

	/* Compute the line vector */
	ldr = sgn(r2 - r1); ldc = sgn(c2 - c1);

	/* Ensure that the legal cartesian line is not illegal in hex */
	if (ldr * ldc == -1)
		AERROR("not a legal line");

	/* Normalize line vector same as motion vector if it is opposite/0 */
	if ((ldr == 0 && ldc == 0) || (ldr == -mdr && ldc == -mdc)) {
		r1 = r2;
		c1 = c2;
		ldr = mdr;
		ldc = mdc;
	}

	/* (Don't use r2,c2,r3,c3 from now on) */
	/* (Only use: r1,c1, ldr,ldc, mdr,mdc) */

	/* Ensure that only the player's pieces are in the described line */
	for (i = 0; i < mylen; i++) {
		r = r1 + i * ldr;
		c = c1 + i * ldc;
		if (a->cell[r][c] != player)
			AERROR("moving invalid piece");
	}

	if (ldr == mdr && ldc == mdc) {
		/* Could be pushing; count foreign things to push */
		for (theirlen = 0; ; theirlen++) {
			r = r1 + (mylen + theirlen) * ldr;
			c = c1 + (mylen + theirlen) * ldc;
			if (isoffboard(r, c) || a->cell[r][c] != OTHER(player))
				break;
		}
		if (theirlen == mylen)
			AERROR("same-numbered push");
		if (theirlen > mylen)
			AERROR("outnumbered push");

		/* Examine cell at the end of their chain */
		r = r1 + (mylen + theirlen) * ldr;
		c = c1 + (mylen + theirlen) * ldc;
		if (theirlen == 0 && isoffboard(r, c))
			AERROR("would push your own piece off the board");
		if (!isoffboard(r, c) && a->cell[r][c] != 0)
				AERROR("no gap to push into");

		/* Committed to moving now! */
		result = new_abalone_move(a, player, cmd);

		if (theirlen > 0) {
			if (isoffboard(r, c))
				/* Pushed their pieces off the board */
				result->lostpieces[OTHER(player)]++;
			else
				/* Push their pieces along */
				result->cell[r][c] = OTHER(player);
		}
		/* Move our piece(s) along */
		result->cell[r1][c1] = 0;
		result->cell[r1 + mylen * ldr][c1 + mylen * ldc] = player;
	} else {
		/* Not moving along the line. Check each piece to move. */
		for (i = 0; i < mylen; i++) {
			r = r1 + i * ldr;
			c = c1 + i * ldc;
			if (isoffboard(r + mdr, c + mdc))
				AERROR("would move your piece off board");
			if (a->cell[r + mdr][c + mdc] != 0)
				AERROR("destination space(s) not free");
		}

		/* Committed to moving now! */
		result = new_abalone_move(a, player, cmd);

		/* Pick them all up */
		for (i = 0; i < mylen; i++) {
			r = r1 + i * ldr;
			c = c1 + i * ldc;
			result->cell[r][c] = 0;
		}

		/* Put them all down */
		for (i = 0; i < mylen; i++) {
			r = r1 + i * ldr;
			c = c1 + i * ldc;
			result->cell[r + mdr][c + mdc] = player;
		}
	}

  check:
	/* Check for 3 duplicate board states */
	i = 0;
	for (b = result->previous; b != NULL; b = b->previous)
		if (memcmp(result->cell, b->cell, sizeof result->cell) == 0)
			i++;
	if (i > 3) {
		free_abalone(result);
		AERROR("board state reached more than 3 times");
	}

#if 1
	/* sanity check: can't lose pieces */
	{
		int p1 = 0;
		int p2 = 0;
		for (r = 0; r < 9; r++)
		for (c = 0; c < 9; c++) {
			if (isoffboard(r,c) && result->cell[r][c] != 0) {
			    fprintf(stderr, "%s: offboard cell %c%c nonzero\n",
				cmd, r+'a',c+'1');
			    exit(1);
			}
			switch (result->cell[r][c]) {
			case 1: p1++; break;
			case 2: p2++; break;
			case 0: break;
			default:
			    fprintf(stderr, "%s: cell %c%c bad value %d\n",
				cmd, r+'a',c+'1', result->cell[r][c]);
			    exit(1);
			}
		}

		if (result->lostpieces[1] > 6 || result->lostpieces[1] + p1 != 14)
		{
			fprintf(stderr, "%s: player 1 pieces mismatch %d+%d\n",
				cmd, p1, result->lostpieces[1]);
			exit(1);
		}

		if (result->lostpieces[2] > 6 || result->lostpieces[2] + p2 != 14)
		{
			fprintf(stderr, "%s: player 2 pieces mismatch %d+%d\n",
				cmd, p2, result->lostpieces[2]);
			exit(1);
		}
	}
#endif

	return result;	/* success */
}
