#include <stdio.h>
#include <stdlib.h>
#include "abalone.h"
#include "ai1.h"

/*
 * Based on minimax example code from
 * http://www.cis.temple.edu/~ingargio/cis587/readings/alpha-beta.c
 */

#define INFINITY	999999
#define MAXDEPTH	3	/* increase this number for better AI */

#undef min
#undef max
#define min(x,y) (((x)<(y))?(x):(y))
#define max(x,y) (((x)>(y))?(x):(y))

/* 
 * AI player. Basic rules are:
 * - pieces on edge are dangerous
 * - avoid pushable segments
 * - don't worry about stragglers
 * - try to divide opponent's pieces
 * - prefer to push marbles off if we have lost some
 * - prefer to push marbles off if only need 1 more to win
 */

struct outcome {
	struct abalone  *a;
	int	score;			/* score of this state */
	char 	cmd[12];		/* command to get here */
	int	bestfscore;		/* best score of all futures */
};

static int compute_score(struct abalone *);

/* Distance from centre */
static abalone_cells distance = 
{
              { 0,0,0,0,0, 0,0,0,0 },
             { 0,1,1,1,1,0, 0,0,0 },
            { 0,1,2,2,2,1,0, 0,0 },
           { 0,1,2,3,3,2,1,0, 0 },
         {  0,1,2,3,4,3,2,1,0  },
        { 0, 0,1,2,3,3,2,1,0 },
       { 0,0, 0,1,2,2,2,1,0 },
      { 0,0,0, 0,1,1,1,1,0 },
     { 0,0,0,0, 0,0,0,0,0 }
};

/* Vectors for the 6 directions */
static struct { coord_t dr, dc; } dir[6] = {
	{+1, 0}, { 0,-1}, {-1,-1}, {-1, 0}, { 0,+1}, {+1,+1}
};

/* Structure for iterating through valid moves. */
struct iterator {
	struct abalone *a;
	player_t player;
	coord_t r1, c1;
	signed char len;	/* line length */
	signed char ld;		/* line direction */
	signed char md;		/* move direction */
	char cmd[9];
};

/* Iterate the next valid abalone board state, returning NULL if none left */
struct abalone *
iterator_next(i)
	struct iterator *i;
{
	struct abalone *anext;

	if (i->r1 >= 9)
		return NULL;

	do {
	  if (i->len == -2) {
	    strcpy(i->cmd, "pass");
	    i->len = -1;
	    i->r1 = i->c1 = 0;
	    i->ld = i->md = 0;
	  } else {
	    if (++(i->len) == 3) {
	      i->len = 0;
	      if (++(i->ld) == 6) {
		i->ld = 0;
		if (++(i->md) == 6) {
		  i->md = 0;
		  if (++(i->c1) == 9) {
		    i->c1 = 0;
		    if (++(i->r1) == 9) {
			  return NULL;
		    }
		  }
		}
	      }
	    }
	    i->cmd[0] = i->r1 + 'a';
	    i->cmd[1] = i->c1 + '1';
	    i->cmd[2] = '-';
	    i->cmd[3] = i->r1 + dir[i->ld].dr * i->len + 'a';
	    i->cmd[4] = i->c1 + dir[i->ld].dc * i->len + '1';
	    i->cmd[5] = ',';
	    i->cmd[6] = i->r1 + dir[i->md].dr + 'a';
	    i->cmd[7] = i->c1 + dir[i->md].dc + '1';
	    i->cmd[8] = '\0';
	  }
	} while ((anext = abalone_move(i->a, i->player, i->cmd)) == NULL);
	return anext;
}

/* Initialise a potential move iterator */
void
iterator_init(i, a, p)
	struct iterator *i;
	struct abalone *a;
	player_t p;
{
	i->a = a;
	i->len = -2;
	i->player = p;
}

/* Return the desirability of a particular board outcome for player 1 */
static int
compute_score(a)
	struct abalone *a;
{
	int score, d;
	coord_t r, c;

	score = 0;

	/*
	 * Score +2dist for every piece that is dist places from the edge,
	 * and -2dist for every opponent piece.
	 */
	for (r = 0; r < 9; r++)
	    for (c = 0; c < 9; c++) {
		switch(a->cell[r][c]) {
		case 1:	score += 2*distance[r][c]; break;
		case 2:	score -= 2*distance[r][c]; break;
		}
	    }

	/*
	 * Score +n for every piece with n neighbours.
	 * score an extra 2 points for far neighbours too.
	 */
	for (r = 0; r < 9; r++)
	    for (c = 0; c < 9; c++) {
		int friends = 0;
		player_t p = a->cell[r][c];
		if (p == 0)
			continue;
		for (d = 0; d < 6; d++) {
			coord_t rf = r + dir[d].dr;
			coord_t cf = c + dir[d].dc;
			if (!isoffboard(rf, cf) &&
			     a->cell[rf][cf] == p) {
				friends++;
				rf += dir[d].dr;
				cf += dir[d].dc;
				if (!isoffboard(rf, cf) &&
				     a->cell[rf][cf] == p) {
					friends += 2;
				}
			}
		}
		switch (p) {
		case 1:	score += 1*friends; break;
		case 2:	score -= 1*friends; break;
		}
	    }

	/*
	 * score 50 for each marble gained/lost
	 */
	score += 50 * a->lostpieces[2];
	score -= 50 * a->lostpieces[1];

	/*
	 * score an extra 200 if we can get 5 pieces
	 */
	if (a->lostpieces[2] == 5)	score += 200;
	if (a->lostpieces[1] == 5)	score -= 200;

	/*
	 * score infinity-1 if we have won (1-Infinity if we lose)
	 */
	if (a->lostpieces[2] == 6)	score = INFINITY-1;
	if (a->lostpieces[1] == 6)	score = 1-INFINITY;

	return score;
}

static int
minimax(n, a, b, depth, player, cmd)
	struct abalone *n;
	int a, b;
	int depth;
	player_t player;
	char *cmd;
{
	int nalpha, nbeta;
	struct abalone *r;
	struct iterator i;

	if (depth == 1) {
		printf("[%s", n->prevcmd);
		fflush(stdout);
	}

	/* Check for a leaf node */
	switch (WINNER(n)) {
	case 1: return INFINITY;
	case 2: return -INFINITY;
	}

	/* Use heuristic if we reach maximum depth */
	if (depth == MAXDEPTH)
		return compute_score(n);

	nalpha = -INFINITY;
	nbeta = INFINITY;
	if (player == 2) {	/* min's move */
		iterator_init(&i, n, player); 
		while ((r = iterator_next(&i)) != NULL) {
			int newb = min(b, nbeta);
			int val = minimax(r, a, newb, depth + 1, 
			    OTHER(player), NULL);
			if (cmd != NULL && val < nbeta)
				strcpy(cmd, r->prevcmd);
			nbeta = min(nbeta, val);
			free_abalone(r);
			if (a >= nbeta)
				break;
		}
		if (depth==1) printf("=%d]", nbeta);
		return nbeta;
	} else {
		iterator_init(&i, n, player);
		while ((r = iterator_next(&i)) != NULL) {
			int newa = max(a, nalpha);
			int val = minimax(r, newa, b, depth + 1,
			    OTHER(player), NULL);
			if (cmd != NULL && val > nalpha)
				strcpy(cmd, r->prevcmd);
			nalpha = max(nalpha, val);
			free_abalone(r);
			if (nalpha >= b)
				break;
		}
		if (depth==1) printf("=%d]", nalpha);
		return nalpha;
	}
}

const char*
ai1_playerfn(a, me)
	struct abalone *a;
	player_t me;
{
	static char bestcmd[12] = "forfeit";

	minimax(a, -INFINITY, INFINITY, 0, me, bestcmd);
	return bestcmd;
}
