#include <stdio.h>
#include <unistd.h>
#include <curses.h>
#include "abalone.h"

static void drawboard(struct abalone *);
static void getcmd(struct abalone *, player_t, const char *, char *, size_t);

#define BLUECELL	(A_NORMAL | COLOR_PAIR(1))
#define REDCELL		(A_NORMAL | COLOR_PAIR(2))
#define GREYTEXT	(A_NORMAL | COLOR_PAIR(3))
#define GREYLINE	(A_NORMAL | COLOR_PAIR(3))
#define GREYCELL	(A_NORMAL | COLOR_PAIR(4))
#define HILITE1S	(A_NORMAL | COLOR_PAIR(5))
#define HILITE1T	(A_NORMAL | COLOR_PAIR(6))
#define HILITE2S	(A_NORMAL | COLOR_PAIR(7))
#define HILITE2T	(A_NORMAL | COLOR_PAIR(8))
#define ERRORTEXT	(A_NORMAL | COLOR_PAIR(9))

/* Set up a user interface using curses */
const char*
cuser_playerfn(a, me)
	struct abalone *a;
	player_t me;
{
	static char cmd[80];
	char prompt[40];
	int turnno;
	struct abalone *b;
	
	/* Standard curses startup */
	initscr();
	use_default_colors();
	start_color();
	cbreak();
	noecho();
	nonl();
	intrflush(stdscr, FALSE);
	keypad(stdscr, TRUE);
	mousemask(BUTTON1_CLICKED, NULL);

	/* Initialise colours */
	init_pair(1, COLOR_BLUE, COLOR_WHITE);
	init_pair(2, COLOR_RED, COLOR_WHITE);
	init_pair(3, COLOR_WHITE, -1);
	init_pair(4, -1, COLOR_WHITE);
	init_pair(5, COLOR_BLUE, COLOR_YELLOW);
	init_pair(6, COLOR_BLUE, COLOR_GREEN);
	init_pair(7, COLOR_RED, COLOR_YELLOW);
	init_pair(8, COLOR_RED, COLOR_GREEN);
	init_pair(9, COLOR_RED, -1);

	for (turnno = 0, b = a; b; b = b->previous, turnno++)
		;

	snprintf(prompt, sizeof prompt, "Turn #%d: Player %d> ", turnno, me);
again:
	erase();
	drawboard(a);
	getcmd(a, me, prompt, cmd, sizeof cmd);
	if (cmd[0] == '\0')
		goto again;

	endwin();

	return cmd;
}

/* Draw one cell of the board */
static void
drawcell(r, c, ch, attrs)
	coord_t r, c;
	char ch;
	int attrs;
{
	int x = 7 + c * 2 - r;
	int y = 2 + r;
	attrset(attrs);
	mvaddch(y, x, ch);
	attrset(A_NORMAL);
}

/*
 * Hilight cells of the board based on a partial command string
 * to give feedback as the command is being entered.
 */
static void
highlight(a, cmd)
	struct abalone *a;
	char *cmd;
{
#define rowof(c)	(((c) >= 'a' && (c) <= 'i') ? (c)-'a' : \
			 ((c) >= 'A' && (c) <= 'I') ? (c)-'A' : -1)
#define colof(c)	((c) >= '1' && (c) <= '9' ? (c) - '1' : -1)

	coord_t r1=-1,c1=-1,r2=-1,c2=-1,r3=-1,c3=-1;
	r1 = rowof(cmd[0]);
	c1 = colof(cmd[1]);
	if (r1!=-1 && c1!=-1 && (cmd[2] == ',' || cmd[2] == '-')) {
	    r2 = rowof(cmd[3]);
	    c2 = colof(cmd[4]);
	    if (r2!=-1 && c2!=-1 && cmd[5] == ',') {
		r3 = rowof(cmd[6]);
		c3 = colof(cmd[7]);
	    }
	}
	if (r3 != -1 && c3 != -1 && cmd[8] != '\0')
		return;
#define hiliteS(r,c)	drawcell(r,c,(a->cell[r][c]==1?'1':a->cell[r][c]==2?'2':' '),a->cell[r][c]==1?HILITE1S:HILITE2S)
#define hiliteT(r,c)	drawcell(r,c,(a->cell[r][c]==1?'1':a->cell[r][c]==2?'2':' '),a->cell[r][c]==1?HILITE1T:HILITE2T)

	if (r1 != -1 && c1 != -1) {
		hiliteS(r1,c1);
		if (r2 != -1 && c2 != -1) {
			if (cmd[2] == ',') {
				hiliteT(r2,c2);
				return;
			} 
			hiliteS(r2,c2);
			/* highlight spot in between ... hmmm */
			hiliteS((r2+r1)/2,(c2+c1)/2);
			if (r3 != -1 && c3 != -1) {
				hiliteT(r3,c3);
			}
		}
	}

#undef rowof
#undef colof
#undef hilite
}

/* Redraw all the cells of the board */
static void
redrawcells(a, nexta)
	struct abalone *a;
	struct abalone *nexta;
{
	coord_t r, c;
	int i;
	int turnno;

        for (r = 0; r < 9; r++) {
            for (c = 0; c < 9; c++) {
		if (isoffboard(r, c))
			continue;
                switch (a->cell[r][c]) {
                case 1:
			drawcell(r, c, '1', BLUECELL);
                        break;
                case 2:
			drawcell(r, c, '2', REDCELL);
                        break;
                default:
			drawcell(r, c, ' ', GREYCELL);
                }
            }
	}

	/* Show ejected pieces */
	attrset(A_NORMAL);
	for (i = 0; i < 6; i++) {
		mvaddch(4, 26 + 2 * i, ' ');
		mvaddch(8, 26 + 2 * i, ' ');
	}

	attrset(BLUECELL);
	for (i = 0; i < a->lostpieces[1]; i++)
		mvaddch(4, 26 + 2 * i, '1');
	attrset(REDCELL);
	for (i = 0; i < a->lostpieces[2]; i++)
		mvaddch(8, 26 + 2 * i, '2');

	attrset(A_NORMAL);
	move(6, 26);
	clrtoeol();
	turnno = 0;
	if (a->previous) {
		struct abalone *b;
		for (b = a; b; b = b->previous)
			turnno++;
		printw("last turn #%d, player %d: %s",
				turnno - 1, a->prevplayer, a->prevcmd);
	} else
		printw("(initial turn)");

	move(7, 26);
	clrtoeol();
	if (nexta) {
		printw("this turn #%d, player %d: %s",
				turnno, nexta->prevplayer, nexta->prevcmd);
	}
}


/*
 * Read keystrokes and mouse clicks to read in a command.
 */
static void
getcmd(a, me, prompt, cmd, cmdmax)
	struct abalone *a;
	player_t me;
	const char *prompt;
	char *cmd;
	size_t cmdmax;
{
	struct abalone *a2;
	struct abalone *cur, *curnext;
	int sy, sx;
	coord_t r, c;
	int len;
	int ch;
	MEVENT evt;

	cur = a;
	curnext = NULL;
	mvaddstr(12, 0, prompt);
	getyx(stdscr, sy, sx);

	len = 0;
	cmd[len] = '\0';

	while (1) {

	    redrawcells(cur, curnext);
	    if (cur == a)
		    highlight(a, cmd);
	    else {
		if (curnext != NULL)
		    highlight(cur, curnext->prevcmd);
	    }

	    mvaddnstr(sy, sx, cmd, len);
	    ch = getch();

	    switch (ch) {
	    case '\n':
	    case '\r':
	    case KEY_ENTER:
		move(14, 0);
		clrtoeol();
		a2 = abalone_move(a, me, cmd);
		if (a2 != NULL) {
			free_abalone(a2);
			return;
		}
		attrset(ERRORTEXT);
		printw("%s", abalone_error);
		cur = a;
		curnext = NULL;
		break;
	    case KEY_BACKSPACE:
		if (len > 0) {
			len--;
			mvaddch(sy, sx + len, ' ');
			move(14, 0);
			clrtoeol();
		}
		break;
	    case KEY_MOUSE:
		if (getmouse(&evt) == OK && len < cmdmax - 3) {
			wmouse_trafo(stdscr, &evt.y, &evt.x, FALSE);
			r = evt.y - 2;
			c = (evt.x + r - 7) / 2;
			/* mvprintw(16, 0, "mouse event at %d %d", r,c); */
			if (!isoffboard(r,c)) {
			    if (len == 2)
				cmd[len++] = ',';
			    if (len == 5 && cmd[2] == ',') {
				cmd[2] = '-';
				cmd[len++] = ',';
			    }
			    cmd[len++] = 'a' + r;
			    cmd[len++] = '1' + c;
			}
		}
		break;

	    case KEY_PPAGE:
		if (cur->previous) {
			curnext = cur;
			cur = cur->previous;
		}
		break;
	    case KEY_NPAGE:
		if (cur != a) {
			struct abalone *b;
			curnext = NULL;
			for (b = a; b && b->previous != cur; b = b->previous)
				curnext = b;
			cur = b;
		}
		break;
	    case KEY_END:
		cur = a;
		curnext = NULL;
		break;
	    case KEY_HOME:
		while (cur->previous != NULL) {
			curnext = cur;
			cur = cur->previous;
		}
		break;

	    default:
		if (ch >= 0x20 && ch < 0x7f && len < cmdmax - 1) {
			cmd[len] = ch;
			len++;
		}
		break;
	    }
	    cmd[len] = '\0';

	}
}

/* Draw a labelled abalone board including the state of the lost pieces */
static void
drawboard(a)
	struct abalone *a;
{
        coord_t r,c;
	int x, y;

        attrset(GREYTEXT);

	/* Column labels */
	for (c = 0; c < 9; c++) {
		x = 9 + c * 2 - (c > 4 ? c - 4 : 0);
		y = (c > 4 ? c - 4 : 0);
		mvaddch(y, x, '1' + c);
		mvaddch(y+1, x-1, '/');
	}

	/* Row labels */
	for (r = 0; r < 9; r++) {
		x = (r < 4 ? 4 - r : r - 4);
		y = 2 + r;
		mvaddch(y, x, 'a' + r);
		mvaddch(y, x+1, '-');
	}

	/* Horizontal motion marks */
        for (r = 0; r < 9; r++) {
            for (c = 0; c < 9; c++) {
                if (!isoffboard(r, c) && !isoffboard(r, c+1)) {
			x = 7 + c * 2 - r;
			y = 2 + r;
			mvaddch(y, x+1, ACS_HLINE);
		}
	    }
	}
	attrset(A_NORMAL);
}
